@@ -111,8 +111,6 @@ int _get_XStringSet_length(SEXP x);
 
 SEXP _get_XStringSet_width(SEXP x);
 
-const char *_get_XStringSet_xsbaseclassname(SEXP x);
-
 XStringSet_holder _hold_XStringSet(SEXP x);
 
 int _get_length_from_XStringSet_holder(const XStringSet_holder *x_holder);
@@ -133,9 +131,14 @@ void _set_XStringSet_names(
 	SEXP names
 );
 
+SEXP _alloc_XStringSet(
+	const char *element_type,
+	SEXP width
+);
+
 SEXP new_XStringSet_from_CHARACTER(
 	SEXP classname,
-	SEXP element_type,
+	SEXP elementType,
 	SEXP x,
 	SEXP start,
 	SEXP width,
@@ -187,7 +190,7 @@ SEXP fasta_index(
 );
 
 SEXP read_XStringSet_from_fasta_blocks(
-	SEXP seqlength,
+	SEXP seqlengths,
 	SEXP filexp_list,
 	SEXP nrec_list,
 	SEXP offset_list,
@@ -202,7 +205,7 @@ SEXP write_XStringSet_to_fasta(
 	SEXP lkup
 );
 
-SEXP fastq_geometry(
+SEXP fastq_seqlengths(
 	SEXP filexp_list,
 	SEXP nrec,
 	SEXP skip,
@@ -216,7 +219,8 @@ SEXP read_XStringSet_from_fastq(
 	SEXP seek_first_rec,
 	SEXP use_names,
 	SEXP elementType,
-	SEXP lkup
+	SEXP lkup,
+	SEXP with_qualities
 );
 
 SEXP write_XStringSet_to_fastq(
