@@ -45,8 +45,8 @@ static const R_CallMethodDef callMethods[] = {
 	CALLMETHOD_DEF(fasta_index, 5),
 	CALLMETHOD_DEF(read_XStringSet_from_fasta_blocks, 6),
 	CALLMETHOD_DEF(write_XStringSet_to_fasta, 4),
-	CALLMETHOD_DEF(fastq_geometry, 4),
-	CALLMETHOD_DEF(read_XStringSet_from_fastq, 7),
+	CALLMETHOD_DEF(fastq_seqlengths, 4),
+	CALLMETHOD_DEF(read_XStringSet_from_fastq, 8),
 	CALLMETHOD_DEF(write_XStringSet_to_fastq, 4),
 
 /* letter_frequency.c */
@@ -174,7 +174,6 @@ void R_init_Biostrings(DllInfo *info)
 
 /* XStringSet_class.c */
 	REGISTER_CCALLABLE(_get_XStringSet_length);
-	REGISTER_CCALLABLE(_get_XStringSet_xsbaseclassname);
 	REGISTER_CCALLABLE(_hold_XStringSet);
 	REGISTER_CCALLABLE(_get_length_from_XStringSet_holder);
 	REGISTER_CCALLABLE(_get_elt_from_XStringSet_holder);
