@@ -26,6 +26,25 @@ static int has_prefix(const char *s, const char *prefix)
 	return 1;
 }
 
+/* TODO: Move this to XVector (together with _copy_CHARSXP_to_Chars_holder). */
+static void copy_Chars_holder(Chars_holder *dest, const Chars_holder *src,
+			      const int *lkup, int lkup_length)
+{
+	char *dest_ptr;
+
+	/* dest->ptr is a (const char *) so we need to cast it to (char *)
+	   before we can write to it */
+	dest_ptr = (char *) dest->ptr;
+	if (lkup == NULL) {
+		memcpy(dest_ptr, src->ptr, dest->length);
+	} else {
+		Ocopy_bytes_to_i1i2_with_lkup(0, dest->length - 1,
+			dest_ptr, dest->length,
+			src->ptr, src->length,
+			lkup, lkup_length);
+	}
+	return;
+}
 
 
 /****************************************************************************
@@ -57,19 +76,20 @@ typedef struct fasta_loader {
 } FASTAloader;
 
 /*
- * The FASTAINDEX loader.
+ * The FASTA INDEX loader.
+ * Used in parse_FASTA_file() to build the FASTA index only.
  */
 
-typedef struct fastaindex_loader_ext {
+typedef struct index_fasta_loader_ext {
 	IntAE *recno_buf;
 	LLongAE *offset_buf;
 	CharAEAE *desc_buf;
 	IntAE *seqlength_buf;
-} FASTAINDEX_loaderExt;
+} INDEX_FASTAloaderExt;
 
-static FASTAINDEX_loaderExt new_FASTAINDEX_loaderExt()
+static INDEX_FASTAloaderExt new_INDEX_FASTAloaderExt()
 {
-	FASTAINDEX_loaderExt loader_ext;
+	INDEX_FASTAloaderExt loader_ext;
 
 	loader_ext.recno_buf = new_IntAE(0, 0, 0);
 	loader_ext.offset_buf = new_LLongAE(0, 0, 0);
@@ -78,11 +98,11 @@ static FASTAINDEX_loaderExt new_FASTAINDEX_loaderExt()
 	return loader_ext;
 }
 
-static void FASTAINDEX_load_desc_line(FASTAloader *loader,
-				      int recno, long long int offset,
-				      const Chars_holder *desc_line)
+static void FASTA_INDEX_load_desc_line(FASTAloader *loader,
+				       int recno, long long int offset,
+				       const Chars_holder *desc_line)
 {
-	FASTAINDEX_loaderExt *loader_ext;
+	INDEX_FASTAloaderExt *loader_ext;
 	IntAE *recno_buf;
 	LLongAE *offset_buf;
 	CharAEAE *desc_buf;
@@ -101,9 +121,9 @@ static void FASTAINDEX_load_desc_line(FASTAloader *loader,
 	return;
 }
 
-static void FASTAINDEX_load_empty_seq(FASTAloader *loader)
+static void FASTA_INDEX_load_empty_seq(FASTAloader *loader)
 {
-	FASTAINDEX_loaderExt *loader_ext;
+	INDEX_FASTAloaderExt *loader_ext;
 	IntAE *seqlength_buf;
 
 	loader_ext = loader->ext;
@@ -112,10 +132,10 @@ static void FASTAINDEX_load_empty_seq(FASTAloader *loader)
 	return;
 }
 
-static void FASTAINDEX_load_seq_data(FASTAloader *loader,
+static void FASTA_INDEX_load_seq_data(FASTAloader *loader,
 		const Chars_holder *seq_data)
 {
-	FASTAINDEX_loaderExt *loader_ext;
+	INDEX_FASTAloaderExt *loader_ext;
 	IntAE *seqlength_buf;
 
 	loader_ext = loader->ext;
@@ -125,8 +145,8 @@ static void FASTAINDEX_load_seq_data(FASTAloader *loader,
 	return;
 }
 
-static FASTAloader new_FASTAINDEX_loader(SEXP lkup, int load_descs,
-					 FASTAINDEX_loaderExt *loader_ext)
+static FASTAloader new_FASTAloader_with_INDEX_ext(SEXP lkup, int load_descs,
+		INDEX_FASTAloaderExt *loader_ext)
 {
 	FASTAloader loader;
 
@@ -137,9 +157,9 @@ static FASTAloader new_FASTAINDEX_loader(SEXP lkup, int load_descs,
 		loader.lkup = INTEGER(lkup);
 		loader.lkup_length = LENGTH(lkup);
 	}
-	loader.load_desc_line = load_descs ? &FASTAINDEX_load_desc_line : NULL;
-	loader.load_empty_seq = &FASTAINDEX_load_empty_seq;
-	loader.load_seq_data = &FASTAINDEX_load_seq_data;
+	loader.load_desc_line = load_descs ? &FASTA_INDEX_load_desc_line : NULL;
+	loader.load_empty_seq = &FASTA_INDEX_load_empty_seq;
+	loader.load_seq_data = &FASTA_INDEX_load_seq_data;
 	loader.nrec = 0;
 	loader.ext = loader_ext;
 	return loader;
@@ -152,11 +172,11 @@ static FASTAloader new_FASTAINDEX_loader(SEXP lkup, int load_descs,
 typedef struct fasta_loader_ext {
 	XVectorList_holder ans_holder;
 	Chars_holder ans_elt_holder;
-} FASTA_loaderExt;
+} FASTAloaderExt;
 
-static FASTA_loaderExt new_FASTA_loaderExt(SEXP ans)
+static FASTAloaderExt new_FASTAloaderExt(SEXP ans)
 {
-	FASTA_loaderExt loader_ext;
+	FASTAloaderExt loader_ext;
 
 	loader_ext.ans_holder = hold_XVectorList(ans);
 	return loader_ext;
@@ -164,7 +184,7 @@ static FASTA_loaderExt new_FASTA_loaderExt(SEXP ans)
 
 static void FASTA_load_empty_seq(FASTAloader *loader)
 {
-	FASTA_loaderExt *loader_ext;
+	FASTAloaderExt *loader_ext;
 	Chars_holder *ans_elt_holder;
 
 	loader_ext = loader->ext;
@@ -179,7 +199,7 @@ static void FASTA_load_empty_seq(FASTAloader *loader)
 static void FASTA_load_seq_data(FASTAloader *loader,
 		const Chars_holder *seq_data)
 {
-	FASTA_loaderExt *loader_ext;
+	FASTAloaderExt *loader_ext;
 	Chars_holder *ans_elt_holder;
 
 	loader_ext = loader->ext;
@@ -192,7 +212,7 @@ static void FASTA_load_seq_data(FASTAloader *loader,
 	return;
 }
 
-static FASTAloader new_FASTA_loader(SEXP lkup, FASTA_loaderExt *loader_ext)
+static FASTAloader new_FASTAloader(SEXP lkup, FASTAloaderExt *loader_ext)
 {
 	FASTAloader loader;
 
@@ -402,7 +422,7 @@ SEXP fasta_index(SEXP filexp_list,
 		 SEXP nrec, SEXP skip, SEXP seek_first_rec, SEXP lkup)
 {
 	int nrec0, skip0, seek_rec0, i, recno, old_nrec, new_nrec, k;
-	FASTAINDEX_loaderExt loader_ext;
+	INDEX_FASTAloaderExt loader_ext;
 	FASTAloader loader;
 	IntAE *seqlength_buf, *fileno_buf;
 	SEXP filexp;
@@ -412,8 +432,8 @@ SEXP fasta_index(SEXP filexp_list,
 	nrec0 = INTEGER(nrec)[0];
 	skip0 = INTEGER(skip)[0];
 	seek_rec0 = LOGICAL(seek_first_rec)[0];
-	loader_ext = new_FASTAINDEX_loaderExt();
-	loader = new_FASTAINDEX_loader(lkup, 1, &loader_ext);
+	loader_ext = new_INDEX_FASTAloaderExt();
+	loader = new_FASTAloader_with_INDEX_ext(lkup, 1, &loader_ext);
 	seqlength_buf = loader_ext.seqlength_buf;
 	fileno_buf = new_IntAE(0, 0, 0);
 	for (i = recno = 0; i < LENGTH(filexp_list); i++) {
@@ -445,7 +465,7 @@ SEXP fasta_index(SEXP filexp_list,
 /* --- .Call ENTRY POINT ---
  * "FASTA blocks" are groups of consecutive FASTA records.
  * Args:
- *   seqlength:   Integer vector with 1 element per record to read. The
+ *   seqlengths:  Integer vector with 1 element per record to read. The
  *                elements must be placed in the order that the records are
  *                going to be read.
  *   filexp_list: A list of N "File External Pointers" (see src/io_utils.c in
@@ -468,29 +488,20 @@ SEXP fasta_index(SEXP filexp_list,
  *                inferred from this).
  *   lkup:        Lookup table for encoding the incoming sequence bytes.
  */
-SEXP read_XStringSet_from_fasta_blocks(SEXP seqlength,
+SEXP read_XStringSet_from_fasta_blocks(SEXP seqlengths,
 		SEXP filexp_list, SEXP nrec_list, SEXP offset_list,
 		SEXP elementType, SEXP lkup)
 {
-	const char *element_type;
-	char classname[40];  /* longest string should be "DNAStringSet" */
 	SEXP ans, filexp, nrec, offset;
-	FASTA_loaderExt loader_ext;
+	FASTAloaderExt loader_ext;
 	FASTAloader loader;
 	int i, j, nrec_j, recno;
 	long long int offset_j, ninvalid;
 
-	element_type = CHAR(STRING_ELT(elementType, 0));
-	if (snprintf(classname, sizeof(classname), "%sSet", element_type)
-	    >= sizeof(classname))
-	{
-		error("Biostrings internal error in "
-		      "read_XStringSet_from_fasta_blocks(): "
-		      "'classname' buffer too small");
-	}
-	PROTECT(ans = alloc_XRawList(classname, element_type, seqlength));
-	loader_ext = new_FASTA_loaderExt(ans);
-	loader = new_FASTA_loader(lkup, &loader_ext);
+	PROTECT(ans = _alloc_XStringSet(CHAR(STRING_ELT(elementType, 0)),
+					seqlengths));
+	loader_ext = new_FASTAloaderExt(ans);
+	loader = new_FASTAloader(lkup, &loader_ext);
 	for (i = 0; i < LENGTH(filexp_list); i++) {
 		filexp = VECTOR_ELT(filexp_list, i);
 		nrec = VECTOR_ELT(nrec_list, i);
@@ -594,44 +605,41 @@ typedef struct fastq_loader {
 } FASTQloader;
 
 /*
- * The FASTQGEOM loader keeps only the common width of the sequences
- * (NA if the set of sequences is not rectangular).
+ * The FASTQ SEQLEN loader.
+ * Used in parse_FASTQ_file() to load the read lengths only.
  */
 
-typedef struct fastqgeom_loader_ext {
-	int width;
-} FASTQGEOM_loaderExt;
+typedef struct seqlen_fastq_loader_ext {
+	IntAE *seqlength_buf;
+} SEQLEN_FASTQloaderExt;
 
-static FASTQGEOM_loaderExt new_FASTQGEOM_loaderExt()
+static SEQLEN_FASTQloaderExt new_SEQLEN_FASTQloaderExt()
 {
-	FASTQGEOM_loaderExt loader_ext;
+	SEQLEN_FASTQloaderExt loader_ext;
 
-	loader_ext.width = NA_INTEGER;
+	loader_ext.seqlength_buf = new_IntAE(0, 0, 0);
 	return loader_ext;
 }
 
-static void FASTQGEOM_load_seq(FASTQloader *loader, const Chars_holder *seq)
+static void FASTQ_SEQLEN_load_seq(FASTQloader *loader, const Chars_holder *seq)
 {
-	FASTQGEOM_loaderExt *loader_ext;
+	SEQLEN_FASTQloaderExt *loader_ext;
+	IntAE *seqlength_buf;
 
 	loader_ext = loader->ext;
-	if (loader->nrec == 0) {
-		loader_ext->width = seq->length;
-		return;
-	}
-	if (loader_ext->width == NA_INTEGER
-	 || seq->length == loader_ext->width)
-		return;
-	loader_ext->width = NA_INTEGER;
+	seqlength_buf = loader_ext->seqlength_buf;
+	IntAE_insert_at(seqlength_buf, IntAE_get_nelt(seqlength_buf),
+			seq->length);
 	return;
 }
 
-static FASTQloader new_FASTQGEOM_loader(FASTQGEOM_loaderExt *loader_ext)
+static FASTQloader new_FASTQloader_with_SEQLEN_ext(
+		SEQLEN_FASTQloaderExt *loader_ext)
 {
 	FASTQloader loader;
 
 	loader.load_seqid = NULL;
-	loader.load_seq = FASTQGEOM_load_seq;
+	loader.load_seq = FASTQ_SEQLEN_load_seq;
 	loader.load_qualid = NULL;
 	loader.load_qual = NULL;
 	loader.nrec = 0;
@@ -644,19 +652,21 @@ static FASTQloader new_FASTQGEOM_loader(FASTQGEOM_loaderExt *loader_ext)
  */
 
 typedef struct fastq_loader_ext {
-	CharAEAE *ans_names_buf;
-	XVectorList_holder ans_holder;
+	CharAEAE *seqid_buf;
+	XVectorList_holder seq_holder;
 	const int *lkup;
 	int lkup_length;
-} FASTQ_loaderExt;
+	XVectorList_holder qual_holder;
+} FASTQloaderExt;
 
-static FASTQ_loaderExt new_FASTQ_loaderExt(SEXP ans, SEXP lkup)
+static FASTQloaderExt new_FASTQloaderExt(SEXP sequences, SEXP lkup,
+		SEXP qualities)
 {
-	FASTQ_loaderExt loader_ext;
+	FASTQloaderExt loader_ext;
 
-	loader_ext.ans_names_buf =
-		new_CharAEAE(_get_XStringSet_length(ans), 0);
-	loader_ext.ans_holder = hold_XVectorList(ans);
+	loader_ext.seqid_buf =
+		new_CharAEAE(_get_XStringSet_length(sequences), 0);
+	loader_ext.seq_holder = hold_XVectorList(sequences);
 	if (lkup == R_NilValue) {
 		loader_ext.lkup = NULL;
 		loader_ext.lkup_length = 0;
@@ -664,47 +674,66 @@ static FASTQ_loaderExt new_FASTQ_loaderExt(SEXP ans, SEXP lkup)
 		loader_ext.lkup = INTEGER(lkup);
 		loader_ext.lkup_length = LENGTH(lkup);
 	}
+	if (qualities != R_NilValue)
+		loader_ext.qual_holder = hold_XVectorList(qualities);
 	return loader_ext;
 }
 
 static void FASTQ_load_seqid(FASTQloader *loader, const Chars_holder *seqid)
 {
-	FASTQ_loaderExt *loader_ext;
-	CharAEAE *ans_names_buf;
+	FASTQloaderExt *loader_ext;
+	CharAEAE *seqid_buf;
 
 	loader_ext = loader->ext;
-	ans_names_buf = loader_ext->ans_names_buf;
+	seqid_buf = loader_ext->seqid_buf;
 	// This works only because seqid->ptr is nul-terminated!
-	CharAEAE_append_string(ans_names_buf, seqid->ptr);
+	CharAEAE_append_string(seqid_buf, seqid->ptr);
 	return;
 }
 
 static void FASTQ_load_seq(FASTQloader *loader, const Chars_holder *seq)
 {
-	FASTQ_loaderExt *loader_ext;
-	Chars_holder ans_elt_holder;
+	FASTQloaderExt *loader_ext;
+	Chars_holder seq_elt_holder;
 
 	loader_ext = loader->ext;
-	ans_elt_holder = get_elt_from_XRawList_holder(&(loader_ext->ans_holder),
-						loader->nrec);
-	/* ans_elt_holder.ptr is a const char * so we need to cast it to
-	   char * before we can write to it */
-	Ocopy_bytes_to_i1i2_with_lkup(0, ans_elt_holder.length - 1,
-		(char *) ans_elt_holder.ptr, ans_elt_holder.length,
-		seq->ptr, seq->length,
-		loader_ext->lkup, loader_ext->lkup_length);
+	seq_elt_holder = get_elt_from_XRawList_holder(
+					&(loader_ext->seq_holder),
+					loader->nrec);
+	copy_Chars_holder(&seq_elt_holder, seq,
+			  loader_ext->lkup, loader_ext->lkup_length);
 	return;
 }
 
-static FASTQloader new_FASTQ_loader(int load_seqids,
-				    FASTQ_loaderExt *loader_ext)
+static void FASTQ_load_qual(FASTQloader *loader, const Chars_holder *qual)
+{
+	FASTQloaderExt *loader_ext;
+	Chars_holder qual_elt_holder;
+
+	loader_ext = loader->ext;
+	qual_elt_holder = get_elt_from_XRawList_holder(
+					&(loader_ext->qual_holder),
+					loader->nrec);
+	copy_Chars_holder(&qual_elt_holder, qual, NULL, 0);
+	return;
+}
+
+static FASTQloader new_FASTQloader(int load_seqids, int load_quals,
+				   FASTQloaderExt *loader_ext)
 {
 	FASTQloader loader;
 
 	loader.load_seqid = load_seqids ? &FASTQ_load_seqid : NULL;
 	loader.load_seq = FASTQ_load_seq;
-	loader.load_qualid = NULL;
-	loader.load_qual = NULL;
+	if (load_quals) {
+		/* Quality ids are always ignored for now. */
+		//loader.load_qualid = &FASTQ_load_qualid;
+		loader.load_qualid = NULL;
+		loader.load_qual = &FASTQ_load_qual;
+	} else {
+		loader.load_qualid = NULL;
+		loader.load_qual = NULL;
+	}
 	loader.nrec = 0;
 	loader.ext = loader_ext;
 	return loader;
@@ -719,7 +748,7 @@ static const char *parse_FASTQ_file(SEXP filexp,
 {
 	int lineno, EOL_in_buf, EOL_in_prev_buf, ret_code,
 	    FASTQ_line1_markup_length, FASTQ_line3_markup_length,
-	    lineinrecno, load_rec;
+	    lineinrecno, load_rec, seq_len;
 	char buf[IOBUF_SIZE];
 	Chars_holder data;
 
@@ -779,8 +808,11 @@ static const char *parse_FASTQ_file(SEXP filexp,
 			}
 		    break;
 		    case 2:
-			if (load_rec && loader->load_seq != NULL)
-				loader->load_seq(loader, &data);
+			if (load_rec) {
+				seq_len = data.length;
+				if  (loader->load_seq != NULL)
+					loader->load_seq(loader, &data);
+			}
 		    break;
 		    case 3:
 			if (!has_prefix(buf, FASTQ_line3_markup)) {
@@ -796,8 +828,18 @@ static const char *parse_FASTQ_file(SEXP filexp,
 			}
 		    break;
 		    case 4:
-			if (load_rec && loader->load_qual != NULL)
-				loader->load_qual(loader, &data);
+			if (load_rec) {
+				if (data.length != seq_len) {
+					snprintf(errmsg_buf, sizeof(errmsg_buf),
+						 "length of quality string "
+						 "at line %d\n  differs from "
+						 "length of corresponding "
+						 "sequence", lineno);
+					return errmsg_buf;
+				}
+				if (loader->load_qual != NULL)
+					loader->load_qual(loader, &data);
+			}
 			if (load_rec)
 				loader->nrec++;
 			(*recno)++;
@@ -812,79 +854,78 @@ static const char *parse_FASTQ_file(SEXP filexp,
 	return NULL;
 }
 
-/* --- .Call ENTRY POINT --- */
-SEXP fastq_geometry(SEXP filexp_list,
-		SEXP nrec, SEXP skip, SEXP seek_first_rec)
+static SEXP get_fastq_seqlengths(SEXP filexp_list,
+		int nrec, int skip, int seek_first_rec)
 {
-	int nrec0, skip0, seek_rec0, i, recno;
-	FASTQGEOM_loaderExt loader_ext;
+	SEQLEN_FASTQloaderExt loader_ext;
 	FASTQloader loader;
+	int recno, i;
+	SEXP filexp;
 	const char *errmsg;
-	SEXP filexp, ans;
 
-	nrec0 = INTEGER(nrec)[0];
-	skip0 = INTEGER(skip)[0];
-	seek_rec0 = LOGICAL(seek_first_rec)[0];
-	loader_ext = new_FASTQGEOM_loaderExt();
-	loader = new_FASTQGEOM_loader(&loader_ext);
+	loader_ext = new_SEQLEN_FASTQloaderExt();
+	loader = new_FASTQloader_with_SEQLEN_ext(&loader_ext);
 	recno = 0;
 	for (i = 0; i < LENGTH(filexp_list); i++) {
 		filexp = VECTOR_ELT(filexp_list, i);
-		errmsg = parse_FASTQ_file(filexp, nrec0, skip0, seek_rec0,
+		errmsg = parse_FASTQ_file(filexp, nrec, skip, seek_first_rec,
 					  &loader, &recno);
 		if (errmsg != NULL)
 			error("reading FASTQ file %s: %s",
 			      CHAR(STRING_ELT(GET_NAMES(filexp_list), i)),
 			      errmsg_buf);
 	}
-	PROTECT(ans = NEW_INTEGER(2));
-	INTEGER(ans)[0] = loader.nrec;
-	INTEGER(ans)[1] = loader_ext.width;
-	UNPROTECT(1);
-	return ans;
+	return new_INTEGER_from_IntAE(loader_ext.seqlength_buf);
 }
 
-/* --- .Call ENTRY POINT --- */
+/* --- .Call ENTRY POINT ---
+ * Args:
+ *   filexp_list:     A list of external pointers.
+ *   nrec:            An integer vector of length 1.
+ *   skip:            An integer vector of length 1.
+ *   seek_first_rec:  A logical vector of length 1.
+ */
+SEXP fastq_seqlengths(SEXP filexp_list,
+		SEXP nrec, SEXP skip, SEXP seek_first_rec)
+{
+	int nrec0, skip0, seek_rec0;
+
+	nrec0 = INTEGER(nrec)[0];
+	skip0 = INTEGER(skip)[0];
+	seek_rec0 = LOGICAL(seek_first_rec)[0];
+	return get_fastq_seqlengths(filexp_list, nrec0, skip0, seek_rec0);
+}
+
+/* --- .Call ENTRY POINT ---
+ * Return an XStringSet object if 'with_qualities' is FALSE, or a list of 2
+ * parallel XStringSet objects of the same shape if 'with_qualities' is TRUE.
+ */
 SEXP read_XStringSet_from_fastq(SEXP filexp_list, SEXP nrec, SEXP skip,
 		SEXP seek_first_rec,
-		SEXP use_names, SEXP elementType, SEXP lkup)
+		SEXP use_names, SEXP elementType, SEXP lkup,
+		SEXP with_qualities)
 {
-	int nrec0, skip0, seek_rec0, load_seqids, ans_length, i, recno;
-	SEXP filexp, ans_geom, ans_width, ans, ans_names;
-	const char *element_type;
-	char classname[40];  /* longest string should be "DNAStringSet" */
-	FASTQ_loaderExt loader_ext;
+	int nrec0, skip0, seek_rec0, load_seqids, load_quals, recno, i;
+	SEXP filexp, seqlengths, sequences, seqids, qualities, ans;
+	FASTQloaderExt loader_ext;
 	FASTQloader loader;
 
 	nrec0 = INTEGER(nrec)[0];
 	skip0 = INTEGER(skip)[0];
 	seek_rec0 = LOGICAL(seek_first_rec)[0];
 	load_seqids = LOGICAL(use_names)[0];
-	PROTECT(ans_geom = fastq_geometry(filexp_list, nrec, skip,
-					  seek_first_rec));
-	ans_length = INTEGER(ans_geom)[0];
-	PROTECT(ans_width = NEW_INTEGER(ans_length));
-	if (ans_length != 0) {
-		if (INTEGER(ans_geom)[1] == NA_INTEGER) {
-			UNPROTECT(2);
-			error("read_XStringSet_from_fastq(): FASTQ files with "
-			      "variable sequence lengths are not supported yet");
-		}
-		for (recno = 0; recno < ans_length; recno++)
-			INTEGER(ans_width)[recno] = INTEGER(ans_geom)[1];
+	load_quals = LOGICAL(with_qualities)[0];
+	PROTECT(seqlengths = get_fastq_seqlengths(filexp_list,
+						  nrec0, skip0, seek_rec0));
+	PROTECT(sequences = _alloc_XStringSet(CHAR(STRING_ELT(elementType, 0)),
+					      seqlengths));
+	if (load_quals) {
+		PROTECT(qualities = _alloc_XStringSet("BString", seqlengths));
+	} else {
+		qualities = R_NilValue;
 	}
-	element_type = CHAR(STRING_ELT(elementType, 0));
-	if (snprintf(classname, sizeof(classname), "%sSet", element_type)
-	    >= sizeof(classname))
-	{
-		UNPROTECT(2);
-		error("Biostrings internal error in "
-		      "read_XStringSet_from_fastq(): "
-		      "'classname' buffer too small");
-	}
-	PROTECT(ans = alloc_XRawList(classname, element_type, ans_width));
-	loader_ext = new_FASTQ_loaderExt(ans, lkup);
-	loader = new_FASTQ_loader(load_seqids, &loader_ext);
+	loader_ext = new_FASTQloaderExt(sequences, lkup, qualities);
+	loader = new_FASTQloader(load_seqids, load_quals, &loader_ext);
 	recno = 0;
 	for (i = 0; i < LENGTH(filexp_list); i++) {
 		filexp = VECTOR_ELT(filexp_list, i);
@@ -893,12 +934,19 @@ SEXP read_XStringSet_from_fastq(SEXP filexp_list, SEXP nrec, SEXP skip,
 				 &loader, &recno);
 	}
 	if (load_seqids) {
-		PROTECT(ans_names =
-		    new_CHARACTER_from_CharAEAE(loader_ext.ans_names_buf));
-		_set_XStringSet_names(ans, ans_names);
+		PROTECT(seqids =
+			new_CHARACTER_from_CharAEAE(loader_ext.seqid_buf));
+		_set_XStringSet_names(sequences, seqids);
 		UNPROTECT(1);
 	}
-	UNPROTECT(3);
+	if (!load_quals) {
+		UNPROTECT(2);
+		return sequences;
+	}
+	PROTECT(ans = NEW_LIST(2));
+	SET_ELEMENT(ans, 0, sequences);
+	SET_ELEMENT(ans, 1, qualities);
+	UNPROTECT(4);
 	return ans;
 }
 
